#!/usr/bin/env python
#
# This script optimizes the placement of a Wireless Access point in a
# specific room.
#
# Author: Joshua A Haas
# Date: 2015-04-28

import math,time
import numpy as np
from population import Population
from room import Room
from wall import Wall

WALL_ATTEN = 5
FREQ = 2400000000
LIGHT = 300000000

# table in the corner
LOC1 = (4,36,1)
# sofa farthest from ethernet
LOC2 = (2,23,1)
# armchair next to ethernet
LOC3 = (8,18,1)
# bedroom 1 desk
LOC4 = (21,6,1)
# bedroom 2 desk
LOC5 = (23,3,1)

locs = [LOC1,LOC2,LOC3,LOC4,LOC5]

r = Room()
r.addwall(0,40,32,40)
r.addwall(0,15,0,40)
r.addwall(0,15,18,15)
r.addwall(12,0,12,15)
r.addwall(12,0,32,0)
r.addwall(32,0,32,32)
r.addwall(32,34,32,36)
r.addwall(32,39,32,40)
r.addwall(22,35,32,35)
r.addwall(22,19,22,35)
r.addwall(22,31,32,31)
r.addwall(22,0,22,15)
r.addwall(21,15,23,15)
r.addwall(20,19,24,19)
r.addwall(27,19,32,19)
r.addwall(26,15,32,15)
r.addwall(29,18,29,19)
r.addwall(29,14,29,16)
r.addwall(29,11,32,11)
r.addwall(29,11,29,12)
r.addwall(12,11,15,11)
r.addwall(15,11,15,12)
r.addwall(15,14,15,15)

def main():
  
  # Define mins and maxs
  
  # Setup population
  mins = np.array([0,0])
  maxs = np.array([32,40])
  p = Population(fitness,cons,mins,maxs,pop_size=1000)
  p.randpop()
  
  iters = 0
  t1 = time.time()
  
  # run for up to 10000 generations, or until convergence
  # print current best every 10 generations
  for i in xrange(0,int(1e3)):
    add = p.evolve(maxiter=10,tol=1e-12,hist=100,mingen=100)
    iters += add
    m = p.getmin()
    x1 = m[0][0]
    x2 = m[0][1]
    y = m[1]
    print str(iters).zfill(5)+': ['+str(x1)+','+str(x2)+'] = '+str(y)
    if add<10:
      break
  
  # print final results
  print '\n'+str(iters)+' generations in '+str(time.time()-t1)+' secs'
  print '\nCalced: ['+str(round(x1,6))+','+str(round(x2,6))+'] = '+str(round(y,6))

# Constraint function to keep members inside the house
def cons(genes):
  
  x = genes[0]
  y = genes[1]
  
  # Bounding box of the layout
  if(x<=0 or x>=32 or y<=0 or y>=40):
    return False
  
  # Do not include blank area in bottom left
  if(x<=12 and y<=15):
    return False
  
  # Do not include bathroom or utility room
  if(x>=22 and x<=32 and y>=19 and y<=31):
    return False
  
  # Do not allow the point to be closer than 1 ft to any loc
  for loc in locs:
    wall = Wall(x,y,loc[0],loc[1])
    if(wall.length<1):
      return False
  
  return True  

# Fitness function to determine optimal placement given weighted locations
def fitness(genes):
  
  total = 0
  for loc in locs:
    wall = Wall(genes[0],genes[1],loc[0],loc[1])
    
    fspl = 20*math.log(4*math.pi*wall.length*0.3048*FREQ/LIGHT,10)
    amp = fspl + WALL_ATTEN*r.num_intersects(wall)
    
    # below 20 db gives no improvement
    amp = max(amp,20)
    
    total += loc[2]*amp
  
  return total/len(locs)

if __name__ == '__main__':
  main()
