#!/usr/bin/env python
#
# The Population class implements a real-valued genetic algorithm
# for finding the minimum of a function.
#
# Author: Joshua A Haas
# Data: 2014-12-01

import time,random
import numpy as np

class Population:
    
  def __init__(self,fit,con,mins,maxs,
      pop=None,pop_size=1000,kill_rate=0.5,mut_rate=0.25):
    """Create a new population for genetic optimization"""
    
    # Set required args
    self.fit_fun = fit
    self.con_fun = con
    self.mins = mins
    self.maxs = maxs
    
    # Set optional args
    self.pop = pop
    self.pop_size = pop_size
    self.kill_rate = kill_rate
    self.mut_rate = mut_rate
    
    # Other setup
    self.genes = len(mins)
    self.hist = np.array([])
    
    if self.pop is None:
      self.pop = np.zeros((self.pop_size,self.genes))
    
    self.fitness = np.zeros(self.pop_size)
    
    self.checkparams()
    
    random.seed(time.time())
  
  def checkparams(self):
    """test all parameters for validity"""
    
    # Type assertions
    assert callable(self.fit_fun), 'Parameter fit must be function'
    assert callable(self.con_fun), 'Parameter con must be function'
    assert isinstance(self.mins,np.ndarray), 'Parameter mins must be numpy.ndarray'
    assert isinstance(self.maxs,np.ndarray), 'Parameter maxs must be numpy.ndarray'
    assert isinstance(self.pop,np.ndarray), 'Parameter pop must be numpy.ndarray'
    assert isinstance(self.pop_size,int), 'Parameter pop_size must be int'
    assert isinstance(self.kill_rate,(int,float)), 'Parameter kill_rate must be int or float'
    assert isinstance(self.mut_rate,(int,float)), 'Parameter mut_rate must be int or float'
    
    # Individual assertions
    assert len(self.mins.shape)==1, 'Parameter mins must be row vector'
    assert len(self.maxs.shape)==1, 'Parameter maxs must be row vector'
    assert len(self.pop)==self.pop_size, 'Parameter pop must be of length pop_size'
    assert self.pop_size>0, 'Parameter pop_size must be greater than 0'
    assert (self.kill_rate>=0) and (self.kill_rate<1), 'Parameter kill_rate must be in [0,1)'
    assert (self.mut_rate>=0) and (self.mut_rate<1), 'Parameter mut_rate must be in [0,1)'
    
    # Dependant assertions
    assert self.pop.shape[0]==self.pop_size, 'Parameter pop must have pop_size rows'
    assert self.pop.shape[1]==self.genes, 'Parameter pop must have genes cols'
    assert self.mins.shape==self.maxs.shape, 'Parameters mins and maxs must be same shape'
    assert all(self.maxs>self.mins), 'Parameter maxs must be greater than mins'
  
  def randpop(self):
    """generate a random population"""

    for r in xrange(0,self.pop_size):
      self.pop[r,:] = self.getrandmem()
    self.updatefit()

  def getrandmem(self):
    """generate a random member using uniform dist"""
    
    feas = False
    mem = np.zeros(self.genes)
    while not feas:
      for g in xrange(0,self.genes):
        mem[g] = random.uniform(self.mins[g],self.maxs[g])
      feas = self.isfeasible(mem)
    return mem

  def isfeasible(self,mem):
    """check if a member is feasible"""
    
    return ((self.con_fun(mem)) and
      all(mem>self.mins) and
      all(mem<self.maxs))

  def updatefit(self,ind=None):
    """update fitness values"""
  
    if ind is None:
      ind = xrange(0,self.pop_size)
    for i in ind:
      self.fitness[i] = self.fit_fun(self.pop[i,:])

  def evolve(self,maxiter=1000,converge=True,tol=1e-6,hist=100,mingen=1):
    """evolve the population until convergence or maxiter"""
    
    converged = False
    i = 0
    while ((not converged) or (not converge)) and (i<maxiter):
      i += 1
      dead = self.getkilled()
      self.breedpop(dead)
      self.updatefit(dead)
      if converge:
        converged = self.isconverged(tol,hist,mingen)
    return i
  
  def getkilled(self):
    """return indices of randomly killed"""
    
    killnum = int(self.kill_rate*self.pop_size)
    return self.fitness.argsort()[-killnum:]
  
  def getunfit(self,pcnt):
    """return the indices of the most unfit"""
    
    unfit = int(self.pop_size*pcnt)
    return self.fitness.argsort()[-unfit:]
  
  def getfit(self,pcnt):
    """return the indices of the most fit"""
  
    fit = int(self.pop_size*pcnt)
    return self.fitness.argsort()[:-(fit+1)]
  
  def breedpop(self,dead):
    """breed new members replacing the given indices"""
  
    alive = np.setdiff1d(xrange(0,self.pop_size),dead)
    for d in dead:
      self.pop[d] = self.breedmem(alive)
  
  def breedmem(self,alive):
    """breed a new member from those that are alive"""
  
    parents = random.sample(alive,2)
    p1 = self.pop[parents[0]]
    p2 = self.pop[parents[1]]
    mu = (p1+p2)/2
    sigma = abs(p2-mu)
    mem = np.zeros(self.genes)
    feas = False
    while not feas:
      for g in xrange(0,self.genes):
        if random.uniform(0,1)<self.mut_rate:
          mem[g] = random.uniform(self.mins[g],self.maxs[g])
        else:
          mem[g] = random.gauss(mu[g],sigma[g])
      feas = self.isfeasible(mem)
    return mem
  
  def isconverged(self,tol,hist,mingen):
    """check for convegence"""
    
    self.hist = np.append(self.hist,self.getmin()[1])
    
    if(len(self.hist)==1):
      return False
    if(len(self.hist)<mingen):
      return False
    if(len(self.hist)>hist):
      self.hist = self.hist[1:]
      
    fit = self.pop[self.getfit(0.5)]
    var = np.var(self.hist)
    return var<tol
  
  def getmin(self):
    """return the best member as (variables,cost)"""
    
    minp = self.fitness.argsort()[0]
    return (self.pop[minp],self.fitness[minp])
